#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
FLASH pulse sequence library.

Calculate the analytical expression of the FLASH pulse sequence signal and
other related quantities.
"""

# ======================================================================
# :: Future Imports
from __future__ import (
    division, absolute_import, print_function, unicode_literals, )

# ======================================================================
# :: Python Standard Library Imports
import os  # Miscellaneous operating system interfaces
# import shutil  # High-level file operations
# import math  # Mathematical functions
# import time  # Time access and conversions
# import datetime  # Basic date and time types
# import operator  # Standard operators as functions
# import collections  # Container datatypes
# import itertools  # Functions creating iterators for efficient looping
# import functools  # Higher-order functions and operations on callable objects
# import argparse  # Parser for command-line options, arguments and subcommands
# import subprocess  # Subprocess management
# import multiprocessing  # Process-based parallelism
# import csv  # CSV File Reading and Writing [CSV: Comma-Separated Values]
# import json  # JSON encoder and decoder [JSON: JavaScript Object Notation]
# import warnings  # Warning control
# import unittest  # Unit testing framework
import doctest  # Test interactive Python examples
import pickle  # Python object serialization

# :: External Imports
import numpy as np  # NumPy (multidimensional numerical arrays library)
import scipy as sp  # SciPy (signal and image processing library)
# import matplotlib as mpl  # Matplotlib (2D/3D plotting library)
import sympy as sym  # SymPy (symbolic CAS library)
# import PIL  # Python Image Library (image manipulation toolkit)
# import SimpleITK as sitk  # Image ToolKit Wrapper
# import nibabel as nib  # NiBabel (NeuroImaging I/O Library)
# import nipy  # NiPy (NeuroImaging in Python)
# import nipype  # NiPype (NiPy Pipelines and Interfaces)
import flyingcircus as fc  # Everything you always wanted to have in Python.*

# :: External Imports Submodules

# :: External Imports Submodules
# import scipy.optimize  # SciPy: Optimization Algorithms
# import scipy.integrate  # SciPy: Integrations facilities
# import scipy.constants  # SciPy: Mathematical and Physical Constants
# import scipy.ndimage  # SciPy: ND-image Manipulation
import scipy.constants  # SciPy: Constants

# :: Internal Imports
import pymrt as mrt
# import pymrt.modules.base
# import pymrt.modules.plot as pmp

# :: Local Imports
from sympy import pi, exp, sin, cos, tan
from pymrt import INFO, PATH
from pymrt import VERB_LVL, D_VERB_LVL, VERB_LVL_NAMES
from pymrt import elapsed, report, run_doctests
from pymrt import msg, dbg
from pymrt.config import CFG


# ======================================================================
def signal(
        m0,
        fa,
        tr,
        t1,
        te,
        t2s,
        eta_fa=1,
        eta_m0=1):
    """
    The FLASH (a.k.a. GRE, TFL, SPGR) signal expression:
    
    s = m0 sin(fa) exp(-te/t2s) (1 - exp(-tr/t1)) / (1 - cos(fa) exp(-tr/t1))
    
    .. math::
        s = \\eta_{m_0} m_0 \\sin(\\eta_\\alpha \\alpha)
        e^{-\\frac{T_E}{T_2^*}}
        \\frac{1 - e^{-\\frac{T_R}{T_1}}}
        {1 - \\cos(\\eta_\\alpha \\alpha) e^{-\\frac{T_R}{T_1}}}

    where
    :math:`m_0` is the spin density,
    :math:`\\eta_{m_0}` is an the receive (spin density) efficiency
    (proportional to the coil receive field :math:`B_1^-`),
    :math:`\\alpha` is the flip angle of the RF excitation,
    :math:`\\eta_\\alpha` is the transmit (flip angle) efficiency
    (proportional to the coil transmit field :math:`B_1^+`),
    :math:`T_E` is the echo time,
    :math:`T_2^*` is the reduced transverse relaxation time,
    :math:`T_R` is the repetition time, and
    :math:`T_1` is the longitudinal relaxation time.

    Args:
        m0 (int|float|np.ndarray): The bulk magnetization M0 in arb. units.
            This includes both spin density and all additional experimental
            factors (coil contribution, electronics calibration, etc.).
        fa (int|float|np.ndarray): The flip angle in rad.
        tr (int|float|np.ndarray): The repetition time in time units.
            Units must be the same as `t1`.
        t1 (int|float|np.ndarray): The longitudinal relaxation in time units.
            Units must be the same as `tr`.
        te (int|float|np.ndarray): The echo time in time units.
            Units must be the same as `t2s`.
        t2s (int|float|np.ndarray): The transverse relaxation in time units.
            Units must be the same as `te`.
        eta_fa (int|float|np.ndarray): The flip angle efficiency in one units.
            This is proportional to the coil transmit field :math:`B_1^+`.
        eta_m0 (int|float|np.ndarray): The spin density efficiency in one
        units.
            This is proportional to the coil receive field :math:`B_1^-`.

    Returns:
        s (float|np.ndarray): The signal expression.
    """
    return eta_m0 * m0 * sin(fa * eta_fa) * exp(-te / t2s) * \
           (1.0 - exp(-tr / t1)) / (1.0 - cos(fa * eta_fa) * exp(-tr / t1))


# ======================================================================
def _eq_expr(*eqs, expr=lambda x: x):
    return sym.Eq(expr([eq.lhs for eq in eqs]), expr([eq.rhs for eq in eqs]))


# ======================================================================
def _simplify_cos(expr):
    return sym.FU['TR5'](expr.expand().trigsimp()).ratsimp()


# ======================================================================
def _prepare_triple_flash_approx(use_cache=CFG['use_cache']):
    """Solve the combination of FLASH images analytically."""

    cache_filepath = os.path.join(
        PATH['cache'], 'flash_triple_approx.cache')
    if not os.path.isfile(cache_filepath) or not use_cache:
        print('Solving Triple FLASH Approx equations. May take some time.')
        print('Caching results: {}'.format(use_cache))

        s, m0, fa, tr, t1, te, t2s, eta_fa, eta_m0 = sym.symbols(
            's m0 fa tr t1 te t2s eta_fa eta_m0', positive=True)
        s1, s2, s3, tr1, tr2, tr3, fa1, fa2, fa3 = sym.symbols(
            's1, s2, s3, tr1 tr2 tr3 fa1 fa2 fa3', positive=True)
        n = sym.symbols('n')
        eq = sym.Eq(s, signal(m0, fa, tr, t1, te, t2s, eta_fa, eta_m0))
        eq_1 = eq.subs({s: s1, fa: fa1, tr: tr1})
        eq_2 = eq.subs({s: s2, fa: fa2, tr: tr2})
        eq_3 = eq.subs({s: s3, fa: fa3, tr: tr3})

        def ratio_expr(x):
            return x[0] / x[1]

        eq_r21 = _eq_expr(eq_2, eq_1, expr=ratio_expr).expand().trigsimp()
        eq_r31 = _eq_expr(eq_3, eq_1, expr=ratio_expr).expand().trigsimp()

        # tr1, tr2, tr3 << t1 approximation
        # fa1, fa2, fa3 ~ 0 approximation
        print('\n', 'Approx: Small FA, Short TR')
        eq_ar21_ = eq_r21.copy()
        for tr_ in tr1, tr2, tr3:
            eq_ar21_ = eq_ar21_.subs(
                exp(-tr_ / t1),
                exp(-tr_ / t1).series(tr_ / t1, n=2).removeO())
        for fa_ in fa1, fa2, fa3:
            for fa_expr in (sin(eta_fa * fa_), cos(eta_fa * fa_)):
                eq_ar21_ = eq_ar21_.subs(
                    fa_expr,
                    fa_expr.series(eta_fa * fa_, n=3).removeO())

        eq_ar31_ = eq_r31.copy()
        for tr_ in tr1, tr2, tr3:
            eq_ar31_ = eq_ar31_.subs(
                exp(-tr_ / t1),
                exp(-tr_ / t1).series(tr_ / t1, n=2).removeO())
        for fa_ in fa1, fa2, fa3:
            for fa_expr in (sin(eta_fa * fa_), cos(eta_fa * fa_)):
                eq_ar31_ = eq_ar31_.subs(
                    fa_expr,
                    fa_expr.series(eta_fa * fa_, n=3).removeO())

        print(eq_ar21_)
        print(eq_ar31_)
        result = sym.nonlinsolve(
            (eq_ar21_, eq_ar31_), (t1, eta_fa))
        for j, exprs in enumerate(result):
            print()
            print('solution: ', j + 1)
            for name, expr in zip(('t1', 'eta_fa'), exprs):
                print(name)
                print(expr)
        t1_ = tuple(exprs[0] for exprs in result)
        eta_fa_ = tuple(exprs[1] for exprs in result)

        pickles = (
            s, m0, fa, tr, t1, te, t2s, eta_fa, eta_m0,
            s1, s2, s3, tr1, tr2, tr3, fa1, fa2, fa3,
            t1_, eta_fa_)
        with open(cache_filepath, 'wb') as cache_file:
            pickle.dump(pickles, cache_file)
    else:
        with open(cache_filepath, 'rb') as cache_file:
            pickles = pickle.load(cache_file)
    print(pickles)
    # result = sym.lambdify(*pickles)
    # return result


# ======================================================================
def _prepare_triple_flash_special1(use_cache=CFG['use_cache']):
    """Solve the combination of FLASH images analytically."""

    cache_filepath = os.path.join(
        PATH['cache'], 'flash_triple_special1.cache')
    if not os.path.isfile(cache_filepath) or not use_cache:
        print('Solving Triple FLASH Special1 equations. May take some time.')
        print('Caching results: {}'.format(use_cache))

        s, m0, fa, tr, t1, te, t2s, eta_fa, eta_m0 = sym.symbols(
            's m0 fa tr t1 te t2s eta_fa eta_m0')
        s1, s2, s3, tr1, tr2, tr3, fa1, fa2, fa3 = sym.symbols(
            's1, s2, s3, tr1 tr2 tr3 fa1 fa2 fa3')
        n = sym.symbols('n')
        eq = sym.Eq(s, signal(m0, fa, tr, t1, te, t2s, eta_fa, eta_m0))
        eq_1 = eq.subs({s: s1, fa: fa1, tr: tr1})
        eq_2 = eq.subs({s: s2, fa: fa2, tr: tr2})
        eq_3 = eq.subs({s: s3, fa: fa3, tr: tr3})

        def ratio_expr(x):
            return x[0] / x[1]

        eq_r21 = _eq_expr(eq_2, eq_1, expr=ratio_expr).expand().trigsimp()
        eq_r31 = _eq_expr(eq_3, eq_1, expr=ratio_expr).expand().trigsimp()

        # tr1, tr2, tr3 << t1 approximation
        # double fa, tr_ratio
        print('\n', 'Special1: Double FA, Short TR')
        double_fa_short_tr = {
            fa1: fa, fa2: 2 * fa, fa3: fa,
            tr1: tr, tr2: tr, tr3: n * tr}

        eq_ar21_ = eq_r21.subs(double_fa_short_tr)
        for tr_ in (tr, n * tr):
            eq_ar21_ = eq_ar21_.subs(
                exp(-tr_ / t1),
                exp(-tr_ / t1).series(tr_ / t1, n=2).removeO())
        eq_ar21_ = sym.FU['TR5'](eq_ar21_.expand().trigsimp())

        eq_ar31_ = eq_r31.subs(double_fa_short_tr)
        for tr_ in (tr, n * tr):
            eq_ar31_ = eq_ar31_.subs(
                exp(-tr_ / t1),
                exp(-tr_ / t1).series(tr_ / t1, n=2).removeO())
        eq_ar31_ = sym.FU['TR5'](eq_ar31_.expand().trigsimp())

        print(eq_ar21_)
        print(eq_ar31_)
        double_fa_short_tr_result = sym.solve(
            (eq_ar21_, eq_ar31_), (t1, cos(eta_fa * fa)))
        for j, exprs in enumerate(double_fa_short_tr_result):
            print('SOLUTION: ', j + 1)
            for name, expr in zip(('t1', 'cos(eta_fa * fa)'), exprs):
                print(name)
                print(expr)

        print('THE END!')

        pickles = (
            s, m0, fa, tr, t1, te, t2s, eta_fa, eta_m0,
            s1, s2, s3, tr1, tr2, tr3, fa1, fa2, fa3,
            double_fa_short_tr_result)
        with open(cache_filepath, 'wb') as cache_file:
            pickle.dump(pickles, cache_file)
    else:
        with open(cache_filepath, 'rb') as cache_file:
            pickles = pickle.load(cache_file)
    print(pickles)
    # result = sym.lambdify(*pickles)
    # return result


# # ======================================================================
# def _prepare_triple_flash_special1(use_cache=CFG['use_cache']):
#     """Solve the combination of FLASH images analytically."""
#
#     cache_filepath = os.path.join(
#         PATH['cache'], 'flash_triple_special1.cache')
#     if not os.path.isfile(cache_filepath) or not use_cache:
#         print('Solving Triple FLASH special1 equations. May take some time.')
#         print('Caching results: {}'.format(use_cache))
#
#         s, m0, fa, tr, t1, te, t2s, eta_fa, eta_m0 = sym.symbols(
#             's m0 fa tr t1 te t2s eta_fa eta_m0')
#         s1, s2, s3, tr1, tr2, tr3, fa1, fa2, fa3 = sym.symbols(
#             's1, s2, s3, tr1 tr2 tr3 fa1 fa2 fa3')
#         n = sym.symbols('n')
#         eq = sym.Eq(s, signal(m0, fa, tr, t1, te, t2s, eta_fa, eta_m0))
#         eq_1 = eq.subs({s: s1, fa: fa1, tr: tr1})
#         eq_2 = eq.subs({s: s2, fa: fa2, tr: tr2})
#         eq_3 = eq.subs({s: s3, fa: fa3, tr: tr3})
#
#         def ratio_expr(x):
#             return x[0] / x[1]
#
#         eq_r21 = _eq_expr(eq_2, eq_1, expr=ratio_expr).expand().trigsimp()
#         eq_r31 = _eq_expr(eq_3, eq_1, expr=ratio_expr).expand().trigsimp()
#
#         # tr1, tr2, tr3 << t1 approximation
#         # double fa, tr_ratio
#         print('\n', ': Double FA, Short TR')
#         double_fa_short_tr = {
#             fa1: fa, fa2: 2 * fa, fa3: fa,
#             tr1: tr, tr2: tr, tr3: n * tr}
#
#         eq_ar21_ = eq_r21.subs(double_fa_short_tr)
#         for tr_ in (tr, n * tr):
#             eq_ar21_ = eq_ar21_.subs(
#                 exp(-tr_ / t1),
#                 exp(-tr_ / t1).series(tr_ / t1, n=2).removeO())
#         eq_ar21_ = sym.FU['TR5'](eq_ar21_.expand().trigsimp())
#
#         eq_ar31_ = eq_r31.subs(double_fa_short_tr)
#         for tr_ in (tr, n * tr):
#             eq_ar31_ = eq_ar31_.subs(
#                 exp(-tr_ / t1),
#                 exp(-tr_ / t1).series(tr_ / t1, n=2).removeO())
#         eq_ar31_ = sym.FU['TR5'](eq_ar31_.expand().trigsimp())
#
#         print(eq_ar21_)
#         print(eq_ar31_)
#         double_fa_short_tr_result = sym.solve(
#             (eq_ar21_, eq_ar31_), (t1, cos(eta_fa * fa)))
#         for j, exprs in enumerate(double_fa_short_tr_result):
#             print('SOLUTION: ', j + 1)
#             for name, expr in zip(('t1', 'cos(eta_fa * fa)'), exprs):
#                 print(name)
#                 print(expr)
#
#         quit()
#
#         # tr1, tr2, tr3 << t1 approximation
#         # half fa, tr_ratio
#         print('\n', ': Half FA, Short TR')
#         half_fa_short_tr = {
#             fa1: fa, fa2: 2 * fa, fa3: 2 * fa,
#             tr1: tr, tr2: tr, tr3: n * tr}
#
#         eq_ar21_ = eq_r21.subs(half_fa_short_tr)
#         for tr_ in (tr, n * tr):
#             eq_ar21_ = eq_ar21_.subs(
#                 exp(-tr_ / t1),
#                 exp(-tr_ / t1).series(tr_ / t1, n=2).removeO())
#         eq_ar21_ = sym.FU['TR5'](eq_ar21_.expand().trigsimp())
#
#         eq_ar31_ = eq_r31.subs(half_fa_short_tr)
#         for tr_ in (tr, n * tr):
#             eq_ar31_ = eq_ar31_.subs(
#                 exp(-tr_ / t1),
#                 exp(-tr_ / t1).series(tr_ / t1, n=2).removeO())
#         eq_ar31_ = sym.FU['TR5'](eq_ar31_.expand().trigsimp())
#
#         print(eq_ar21_)
#         print(eq_ar31_)
#         # half_fa_short_tr_result = sym.solve(
#         #     (eq_ar21_, eq_ar31_), (t1, cos(eta_fa * fa)))
#         # for j, exprs in enumerate(half_fa_short_tr_result):
#         #     print()
#         #     print('solution: ', j + 1)
#         #     for name, expr in zip(('t1', 'cos(eta_fa * fa)'), exprs):
#         #         print(name)
#         #         print(expr)
#
#         # tr1, tr2, tr3 << t1 approximation
#         # fa1, fa2, fa3 ~ 0 approximation
#         print('\n', ': Small FA, Short TR')
#         eq_ar21_ = eq_r21.copy()
#         for tr_ in tr1, tr2, tr3:
#             eq_ar21_ = eq_ar21_.subs(
#                 exp(-tr_ / t1),
#                 exp(-tr_ / t1).series(tr_ / t1, n=2).removeO())
#         for fa_ in fa1, fa2, fa3:
#             for fa_expr in (sin(eta_fa * fa_), cos(eta_fa * fa_)):
#                 eq_ar21_ = eq_ar21_.subs(
#                     fa_expr,
#                     fa_expr.series(eta_fa * fa_, n=3).removeO())
#
#         eq_ar31_ = eq_r31.copy()
#         for tr_ in tr1, tr2, tr3:
#             eq_ar31_ = eq_ar31_.subs(
#                 exp(-tr_ / t1),
#                 exp(-tr_ / t1).series(tr_ / t1, n=2).removeO())
#         for fa_ in fa1, fa2, fa3:
#             for fa_expr in (sin(eta_fa * fa_), cos(eta_fa * fa_)):
#                 eq_ar31_ = eq_ar31_.subs(
#                     fa_expr,
#                     fa_expr.series(eta_fa * fa_, n=3).removeO())
#
#         print(eq_ar21_)
#         print(eq_ar31_)
#         # small_fa_short_tr_result = sym.solve(
#         #     (eq_ar21_, eq_ar31_), (t1, eta_fa))
#         # for j, exprs in enumerate(small_fa_short_tr_result):
#         #     print()
#         #     print('solution: ', j + 1)
#         #     for name, expr in zip(('t1', 'eta_fa'), exprs):
#         #         print(name)
#         #         print(expr)
#
#         print('THE END!')
#         # quit()
#         # pickles = ()
#         # with open(cache_filepath, 'wb') as cache_file:
#         #     pickle.dump(pickles, cache_file)
#     else:
#         with open(cache_filepath, 'rb') as cache_file:
#             pickles = pickle.load(cache_file)
#     # result = sym.lambdify(*pickles)
#     # return result

# ======================================================================
# :: defines the mp2rage signal expression
# _rho = _prepare_triple_flash_approx()


# ======================================================================
def rotation(
        angle=sym.Symbol('a'),
        axes=(0, 1),
        n_dim=3):
    rot_mat = sym.eye(n_dim)
    rot_mat[axes[0], axes[0]] = cos(angle)
    rot_mat[axes[1], axes[1]] = cos(angle)
    rot_mat[axes[0], axes[1]] = -sin(angle)
    rot_mat[axes[1], axes[0]] = sin(angle)
    return rot_mat


# ======================================================================
def resonance_offset_evolution(
        time=sym.Symbol('t'),
        magnetic_field_variation=sym.Symbol('B_Delta'),
        gamma=sp.constants.physical_constants['proton gyromagn. ratio'][0]):
    return gamma * magnetic_field_variation * time


# ======================================================================
def evolution(
        initial_magnetization,
        flip_angle=sym.Symbol('a'),
        rotation_plane=(1, 2),
        duration=sym.Symbol('t'),
        relaxation_longitudinal=sym.Symbol('R_1'),
        relaxation_transverse=sym.Symbol('R_2'),
        resonance_offset=0,
        equilibrium_magnetization=sym.Symbol('M_eq')):
    decay = rotation(resonance_offset, (0, 1))
    decay[0:2, 0:2] *= exp(-duration * relaxation_transverse)
    decay[-1, -1] *= exp(-duration * relaxation_longitudinal)
    recovery = sym.Matrix(
        [0, 0, equilibrium_magnetization *
         (1 - exp(-duration * relaxation_longitudinal))])
    excitation = rotation(flip_angle, rotation_plane)
    final_magnetization = decay * excitation * initial_magnetization + recovery
    return final_magnetization, excitation


# ======================================================================
def evolution_flash(
        initial_magnetization,
        repetition_time=sym.Symbol('T_R'),
        flip_angle=sym.Symbol('a')):
    n_dim = 3

    relaxation_longitudinal = sym.Symbol('R_1')
    relaxation_transverse = sym.Symbol('R_2')
    resonance_offset = 0
    equilibrium_magnetization = sym.Symbol('M_eq')

    final_magnetization, first_excitation = evolution(
        initial_magnetization, flip_angle, (1, 2), repetition_time,
        relaxation_longitudinal, relaxation_transverse, resonance_offset,
        equilibrium_magnetization)

    # attempt some simplification
    for i in range(n_dim):
        final_magnetization[i] = sym.trigsimp(final_magnetization[i])
    return final_magnetization, first_excitation


# ======================================================================
def steady_state(
        evolution_func,
        *evolution_args,
        **evolution_kws):
    steady_state_magnetization_minus = magnetization('ss')
    final_magnetization, first_excitation = evolution_func(
        steady_state_magnetization_minus,
        *evolution_args, **evolution_kws)
    eqn_steady_state = sym.Eq(
        steady_state_magnetization_minus, final_magnetization)
    steady_state_solution = sym.solve(
        eqn_steady_state, steady_state_magnetization_minus)
    steady_state_magnetization_minus = sym.Matrix(
        [steady_state_solution[item]
         for item in steady_state_magnetization_minus])
    steady_state_magnetization_plus = \
        first_excitation * steady_state_magnetization_minus
    return steady_state_magnetization_minus, steady_state_magnetization_plus


# ======================================================================
def magnetization(
        label='',
        n_dim=3):
    mag_vec = sym.Matrix(
        [sym.Symbol('M_{}_{}'.format(label, i))
         for i in range(n_dim)])
    return mag_vec


# ======================================================================
@np.vectorize
def ernst_calc(
        t1=None,
        tr=None,
        fa=None):
    """
    Calculate optimal T1, TR or FA (given the other two) for FLASH sequence.

    Args:
        t1 (float|np.ndarray|None): Longitudinal relaxation time T1 in ms.
        tr (float|np.ndarray|None): Repetition time TE in ms.
        fa (float|np.ndarray|None): flip angle FA in deg.

    Returns:
        val (float|np.ndarray|None): The value of fulfilling Ernst condition.
            This can be T1, TR or FA, depending on the input argument
            left to None.
            If the input is invalid, this is set to None.
        name (str|None): The label of the result.
            If the input is invalid, this is set to None.
        units (str|None): The units of measurement for the result.
            If the input is invalid, this is set to None.

    Examples:
        >>> ernst_calc(100.0, 30.0)
        (42.198837866408269, 'FA', 'deg')
        >>> ernst_calc(100.0, fa=42)
        (29.686433336996988, 'TR', 'ms')
        >>> ernst_calc(None, 30.0, 42)
        (101.05626250025875, 'T1', 'ms')
    """
    if t1 and tr:
        fa = np.arccos(np.exp(-tr / t1))
        fa = np.rad2deg(fa)
        val, name, units = fa, 'FA', 'deg'
    elif tr and fa:
        fa = np.deg2rad(fa)
        t1 = -tr / np.log(np.cos(fa))
        val, name, units = t1, 'T1', 'ms'
    elif t1 and fa:
        fa = np.deg2rad(fa)
        tr = -t1 * np.log(np.cos(fa))
        val, name, units = tr, 'TR', 'ms'
    else:
        val = name = units = None
    return val, name, units


# ======================================================================
def print_steady_state():
    ss = steady_state(evolution_flash)
    signal = sym.sqrt(sym.trigsimp(ss[1][0] * ss[1][0] + ss[1][1] * ss[1][1]))
    print('\nSteady-State before excitation:')
    print(ss[0])
    print('\nSteady-State after excitation:')
    print(ss[1])
    print('\nSignal expression:')
    print(signal)


# ======================================================================
elapsed(__file__[len(PATH['base']) + 1:])

# ======================================================================
if __name__ == '__main__':
    run_doctests(__doc__)
